# text-based tic-tac-toe with AI opponent
# Aaron Bauer, 2019

import copy

def get_opponent(player):
    if player == "X":
        return "O"
    return "X"

class TicTacToeState():
    def __init__(self, board, turn):
        self.board = board
        self.turn = turn
        self.winner = self.check_winner()
        # self.next_moves = []
        # if self.winner == "":
        #     for i in range(len(self.board)):
        #         for j in range(len(self.board[i])):
        #             if self.board[i][j] == "":
        #                 next_board = copy.deepcopy(self.board)
        #                 next_board[i][j] = turn
        #                 if turn == "X":
        #                     next_turn = "O"
        #                 else:
        #                     next_turn = "X"
        #                 self.next_moves.append(TicTacToeState(next_board, next_turn))

    # while we could recursively assemble the lists of next moves ahead of time in __init__
    # we can avoid the delay when we first run the program by putting it in a separate method
    def get_next_moves(self):
        """returns a list of TicTacToeStates for the set of possible moves from the current state"""
        moves = []
        if self.winner == "":
            for i in range(len(self.board)):
                for j in range(len(self.board[i])):
                    if self.board[i][j] == "":
                        next_board = copy.deepcopy(self.board)
                        next_board[i][j] = self.turn
                        next_turn = get_opponent(self.turn)
                        moves.append(TicTacToeState(next_board, next_turn))
            return moves

    def check_row(self, i):
        """check row i for a winner"""
        return self.board[i][0] == self.board[i][1] == self.board[i][2] and len(self.board[i][0]) > 0

    def check_column(self, i):
        """check column i for a winner"""
        return self.board[0][i] == self.board[1][i] == self.board[2][i] and len(self.board[0][i]) > 0

    def check_winner(self):
        """check each row, column, and diagonal for a winner
           return "X" if X wins, "O" if O wins, and "" if there is no winner yet"""
        for i in range(len(self.board)):
            if self.check_row(i):
                return self.board[i][0]
            if self.check_column(i):
                return self.board[0][i]
        if self.board[0][0] == self.board[1][1] == self.board[2][2] and len(self.board[0][0]) > 0:
            return self.board[0][0]
        if self.board[0][2] == self.board[1][1] == self.board[2][0] and len(self.board[0][2]) > 0:
            return self.board[0][2]
        return ""

    def is_full(self):
        """return True if all spaces have been filled, otherwise False"""
        full = True
        for i in range(len(self.board)):
            for j in range(len(self.board[i])):
                if self.board[i][j] == "":
                    full = False
        return full

    def make_move(self, move):
        """move must be between 0 and 8 and correspond to an open board location
           modifies this state as if self.turn made a mark at the location indicated by move"""
        if move < 0 or move > 8 or self.board[move // 3][move % 3] != "":
            raise ValueError("{} is not a valid move for {}".format(move, self.board))
        self.board[move // 3][move % 3] = self.turn
        self.turn = get_opponent(self.turn)
        self.winner = self.check_winner()

    def get_best_move(self):
        moves = self.get_next_moves()
        if len(moves) == 0:
            return self
        best_move = moves[0]
        best_result = best_move.result(self.turn)
        for move in moves[1:]:
            r = move.result(self.turn)
            if r > best_result:
                best_move = move
                best_result = r
        return best_move

        # # an equivalent, and much more concise, version of the above code:
        # return max(self.get_next_moves(), key=lambda x: x.result(self.turn),
        #            default=self)

    def result(self, player):
        moves = self.get_next_moves()

        # if this state is a win for player, score it as 1
        if self.winner == player:
            return 1

        # if this state is a loss for player (i.e., the winner is player's opponent), score it as a -1
        if self.winner == get_opponent(player):
            return -1

        # if any move from this state leads to a loss for player, also score it a -1
        for move in moves:
            if m.winner == get_opponent(player) and m.winner != player:
                return -1

        # if there are no moves (i.e., this state is a draw), score it a 0
        if len(moves) == 0:
            return 0

        # at this point, we can't determine a score based on this state or it's immediate moves
        # so we recursively score each of the possible successor states
        results = []
        for move in moves:
            results.append(move.result(player))

        # return the max score among the possible moves, so we prioritize possible wins over draws over losses
        return max(results)

    def __repr__(self):
        r = ""
        for row in self.board:
            for cell in row:
                if cell == "":
                    cell = "_"
                r += cell + "   "
            r += "\n"
        return r

def take_turn(state):
    """take a human turn in tic-tac-toe"""
    # display the current state and the labels for choosing a move
    print(state)
    print("""0   1   2
3   4   5
6   7   8""")
    move = input("enter a number to move: ")
    # prompt again for a move until it's a valid input and corresponds to an empty space in the board
    while not move.isnumeric() or len(move) > 1 or not (0 <= int(move) <= 8) or state.board[int(move) // 3][int(move) % 3] != "":
        move = input("enter a valid move: ")
    state.make_move(move)

def play():
    board = [["", "", ""],
             ["", "", ""],
             ["", "", ""]]
    root = TicTacToeState(board, "X")
    while root.winner == "" and not root.is_full():
        take_turn(root)
        if root.winner == "":
            # make AI move
            root = root.get_best_move()
    print(root)
    if root.winner != "":
        print(root.winner, "wins!")
    else:
        print("It's a draw.")
play()
